from allosaurus.am.allosaurus_torch import AllosaurusTorchModel
from allosaurus.am.utils import *
from allosaurus.lm.inventory import Inventory
from allosaurus.lm.unit import write_unit
import json
from argparse import Namespace
from allosaurus.model import get_model_path

def read_am(model_path, inference_config):
    """
    load pretrained acoustic model

    :param model_path: path to the
    :return:
    """

    am_config = Namespace(**json.load(open(str(model_path / 'am_config.json'))))

    assert am_config.model == 'allosaurus', "This project only support allosaurus model"

    model = AllosaurusTorchModel(am_config)

    # load weights
    torch_load(model, str(model_path / 'model.pt'), inference_config.device_id)

    return model

def transfer_am(train_config):
    """
    initialize the acoustic model with a pretrained model for fine-tuning

    :param model_path: path to the
    :return:
    """

    pretrained_model_path = get_model_path(train_config.pretrained_model)

    am_config = Namespace(**json.load(open(str(pretrained_model_path / 'am_config.json'))))

    assert am_config.model == 'allosaurus', "This project only support allosaurus model"

    # load inventory
    inventory = Inventory(pretrained_model_path)

    # get unit_mask which maps the full phone inventory to the target phone inventory
    unit_mask = inventory.get_mask(train_config.lang, approximation=True)

    # reset the new phone_size
    am_config.phone_size = len(unit_mask.target_unit)

    model = AllosaurusTorchModel(am_config)

    # load the pretrained model and setup the phone_layer with correct weights
    torch_load(model, str(pretrained_model_path / 'model.pt'), train_config.device_id, unit_mask)

    # update new model
    new_model = train_config.new_model

    # get its path
    model_path = get_model_path(new_model)

    # overwrite old am_config
    new_am_config_json = vars(am_config)
    json.dump(new_am_config_json, open(str(model_path / 'am_config.json'), 'w'), indent=4)

    # overwrite old phones
    write_unit(unit_mask.target_unit, model_path / 'phone.txt')

    # overwrite old model
    torch_save(model, model_path / 'model.pt')

    return model


import numpy as np

def read_unit(unit_path):
    # load unit from units.txt
    # units.txt should start from index 1 (because ctc blank is taking the 0 index)

    unit_to_id = dict()

    unit_to_id['<blk>'] = 0

    idx = 0

    for line in open(str(unit_path), 'r', encoding='utf-8'):
        fields = line.strip().split()

        assert len(fields) < 3, " each line should contain at most two field separated by space."

        # this is simple format
        if len(fields) == 1:
            unit = fields[0]
            idx += 1
        else:
            # this is kaldi format
            unit = fields[0]
            idx = int(fields[1])

        assert unit not in unit_to_id, "there are duplicate phones."
        unit_to_id[unit] = idx

    unit = Unit(unit_to_id)
    return unit

def write_unit(unit, unit_path, format='kaldi'):
    """
    dump units to file

    :param unit:
    :param unit_path:
    :return:
    """

    # unit can be either kaldi format (each line contain two field phone and index) or
    # the simple format (each line contains only phone)
    assert format in ['kaldi', 'simple']

    w = open(str(unit_path), 'w', encoding='utf-8')
    for i in range(1, len(unit.id_to_unit)):
        u = unit.id_to_unit[i]

        if format == 'kaldi':
            w.write(u+' '+str(i)+'\n')
        else:
            w.write(u+'\n')

    w.close()

class Unit:

    def __init__(self, unit_to_id):
        """
        Unit manages bidirectional mapping from unit to id and id to unit
        both are dict

        :param unit_to_id:
        """

        self.unit_to_id = unit_to_id
        self.id_to_unit = {}

        assert '<blk>' in self.unit_to_id and self.unit_to_id['<blk>'] == 0

        for unit, idx in self.unit_to_id.items():
            self.id_to_unit[idx] = unit

    def __str__(self):
        return '<Unit: ' + str(len(self.unit_to_id)) + ' elems>'

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, idx):
        return self.id_to_unit[idx]

    def __len__(self):
        return len(self.id_to_unit)

    def __contains__(self, unit):
        if unit == ' ':
            unit = '<space>'

        return unit in self.unit_to_id

    def get_id(self, unit):

        # handle special units
        if unit == ' ':
            unit = '<space>'

        assert unit in self.unit_to_id, 'unit '+unit+'is not in '+str(self.unit_to_id)
        return self.unit_to_id[unit]

    def get_ids(self, units):
        """
        get index for a word list
        :param words:
        :return:
        """

        return [self.get_id(unit) for unit in units]

    def get_unit(self, id):
        assert id >= 0 and id in self.id_to_unit

        unit = self.id_to_unit[id]

        # handle special units
        if unit == '<space>':
            unit = ' '

        return unit

    def get_units(self, ids):
        """
        get unit from ids

        :param ids: elem_id list
        :return: a list of unit
        """

        return [self.get_unit(id) for id in ids]


import argparse
from pathlib import Path
from allosaurus.model import resolve_model_name
from allosaurus.lm.inventory import *
from tqdm import tqdm

def prepare_token(data_path, model, lang_id):

    model_path = Path(__file__).parent.parent / 'pretrained' / model

    #assert model_path.exists(), f"{model} is not a valid model"

    inventory = Inventory(model_path)
    unit = inventory.get_unit(lang_id)

    writer = open(str(data_path / 'token'), 'w', encoding='utf-8')

    for line in tqdm(open(data_path / 'text', 'r', encoding='utf-8').readlines()):
        fields = line.strip().split()
        utt_id = fields[0]

        phones = fields[1:]

        id_lst = unit.get_ids(phones)

        writer.write(utt_id+' '+' '.join(map(str, id_lst))+'\n')

    writer.close()

if __name__ == '__main__':

    parser = argparse.ArgumentParser('allosaurus tool to assign token id for fine-tuning')
    parser.add_argument('--path', required=True, type=str, help='path to the directory containing the text file')
    parser.add_argument('--model', type=str, default='latest', help='specify the model you want to fine-tune')
    parser.add_argument('--lang', type=str, default='epi', help='specify the ISO language id for your target language')

    args = parser.parse_args()
    data_path = Path(args.path)

    text_path = data_path / 'text'

    assert text_path.exists(), "the path directory should contain a text file, please check README.md for details"

    # resolve model's name
    model_name = resolve_model_name(args.model)
    if model_name == "none":
        print("Model ", model_name, " does not exist. Please download this model or use an existing model in list_model")
        exit(0)

    args.model = model_name

    # extract token
    prepare_token(data_path, args.model, args.lang)


from torch.optim import SGD

def read_optimizer(model, train_config):

    assert train_config.optimizer == 'sgd', 'only sgd is supported now, others optimizers would be easier to add though'

    return SGD(model.parameters(), lr=train_config.lr)


from allosaurus.model import get_all_models


if __name__ == '__main__':

    models = get_all_models()

    if len(models) == 0:
        print("No models are available, you can maually download a model with download command or just run inference to download the latest one automatically")
    else:
        print("Available Models")
        for i, model in enumerate(models):
            if i == 0:
                print('-', model.name, "(default)")
            else:
                print('-', model.name)


import torch
import torch.nn as nn


def read_criterion(train_config):

    assert train_config.criterion == 'ctc', 'only ctc criterion is supported now'
    return CTCCriterion(train_config)


class CTCCriterion(nn.Module):
    def __init__(self, train_config):
        super().__init__()
        self.train_config = train_config

        self.logsoftmax = nn.LogSoftmax(dim=2)

        self.criterion = nn.CTCLoss(reduction='sum', zero_infinity=True)

    def forward(self,
                output_tensor: torch.tensor,
                output_lengths: torch.tensor,
                target_tensor: torch.tensor,
                target_lengths: torch.tensor):

        output_tensor = self.logsoftmax(output_tensor).transpose(0,1)
        loss = self.criterion(output_tensor, target_tensor, output_lengths, target_lengths)
        return loss


import numpy
import struct
import functools
import os.path
import gzip
import bz2
from pathlib import Path


class KaldiWriter:

    def __init__(self, path=None, scp=True):
        """
        writer of BOTH ark and scp

        :param path:
        :param scp:
        """

        self.scp = scp

        if path:
            self.open(path)

        self.ark_offset = 0

    def open(self, path):

        # remove scp or ark suffix
        path = str(path)
        if path.endswith(".scp") or path.endswith(".ark"):
            path = path[:-4]

        # mkdir
        p = Path(path).parent
        p.mkdir(parents=True, exist_ok=True)

        self.ark_path = str(path) + '.ark'
        self.scp_path = str(path) + '.scp'

        # delete old files
        if os.path.exists(self.ark_path):
            os.remove(self.ark_path)

        if os.path.exists(self.scp_path):
            os.remove(self.scp_path)

        self.ark_writer = open(self.ark_path, "ab")
        self.scp_writer = open(self.scp_path, "a")
        self.ark_offset = self.ark_writer.tell()

    def write(self, utt_ids, feats):

        # write batch
        if isinstance(utt_ids, list):
            pointers = []

            for utt_id, feat in zip(utt_ids, feats):
                self.ark_offset += write_string(self.ark_writer, utt_id)
                pointers.append("%s:%d" % (self.ark_path, self.ark_offset))
                self.ark_offset += write_matrix(self.ark_writer, feat)

            #self.ark_writer.flush()

            for utt_id, pointer in zip(utt_ids, pointers):
              self.scp_writer.write("%s %s\n" % (utt_id, pointer))

            #self.scp_writer.flush()

        else:
            # write single instance
            utt_id = utt_ids
            feat = feats
            self.ark_offset += write_string(self.ark_writer, utt_id)
            pointer = "%s:%d" % (self.ark_path, self.ark_offset)
            self.ark_offset += write_matrix(self.ark_writer, feat)
            self.scp_writer.write("%s %s\n" % (utt_id, pointer))

    def close(self):
        self.ark_writer.close()
        self.scp_writer.close()


def smart_open(filename, mode='rb', *args, **kwargs):
    '''
    Opens a file "smartly":
      * If the filename has a ".gz" or ".bz2" extension, compression is handled
        automatically;
      * If the file is to be read and does not exist, corresponding files with
        a ".gz" or ".bz2" extension will be attempted.
    '''
    readers = {'.gz': gzip.GzipFile, '.bz2': bz2.BZ2File}
    if 'r' in mode and not os.path.exists(filename):
        for ext in readers:
            if os.path.exists(filename + ext):
                filename += ext
                break
    extension = os.path.splitext(filename)[1]
    return readers.get(extension, open)(filename, mode, *args, **kwargs)


def read_string(f):
    s = ""
    while True:
        c = f.read(1).decode('utf-8')
        if c == "": raise ValueError("EOF encountered while reading a string.")
        if c == " ": return s
        s += c


def read_integer(f):
    n = ord(f.read(1))
    # return reduce(lambda x, y: x * 256 + ord(y), f.read(n)[::-1], 0)
    a = f.read(n)[::-1]
    try:
        return int.from_bytes(a, byteorder='big', signed=False)
    except:
        return functools.reduce(lambda x, y: x * 256 + ord(y), a, 0)
    # return functools.reduce(lambda x, y: x * 256 + ord(y), f.read(n)[::-1].decode('windows-1252'), 0)
    # try:
    # a=f.read(n)[::-1]
    # b=int.from_bytes(a, byteorder='big', signed=False)
    # print(a,type(a),b)
    # return functools.reduce(lambda x, y: x * 256 + ord(y), a[::-1], 0)
    # return functools.reduce(lambda x, y: x * 256 + ord(y), f.read(n)[::-1], 0)
    # except:
    #    return functools.reduce(lambda x, y: x * 256 + ord(y), f.read(n)[::-1].decode('windows-1252'), 0)


def read_compressed_matrix(fd, format):
    """ Read a compressed matrix,
        see: https://github.com/kaldi-asr/kaldi/blob/master/src/matrix/compressed-matrix.h
        methods: CompressedMatrix::Read(...), CompressedMatrix::CopyToMat(...),
    """

    # Format of header 'struct',
    global_header = numpy.dtype([('minvalue', 'float32'), ('range', 'float32'), ('num_rows', 'int32'),
                                 ('num_cols', 'int32')])  # member '.format' is not written,
    per_col_header = numpy.dtype([('percentile_0', 'uint16'), ('percentile_25', 'uint16'), ('percentile_75', 'uint16'),
                                  ('percentile_100', 'uint16')])

    # Read global header,
    globmin, globrange, rows, cols = numpy.frombuffer(fd.read(16), dtype=global_header, count=1)[0]

    # The data is structed as [Colheader, ... , Colheader, Data, Data , .... ]
    #                         {           cols           }{     size         }
    col_headers = numpy.frombuffer(fd.read(cols * 8), dtype=per_col_header, count=cols)
    col_headers = numpy.array(
        [numpy.array([x for x in y]) * globrange * 1.52590218966964e-05 + globmin for y in col_headers],
        dtype=numpy.float32)
    data = numpy.reshape(numpy.frombuffer(fd.read(cols * rows), dtype='uint8', count=cols * rows),
                         newshape=(cols, rows))  # stored as col-major,

    mat = numpy.zeros((cols, rows), dtype='float32')
    p0 = col_headers[:, 0].reshape(-1, 1)
    p25 = col_headers[:, 1].reshape(-1, 1)
    p75 = col_headers[:, 2].reshape(-1, 1)
    p100 = col_headers[:, 3].reshape(-1, 1)
    mask_0_64 = (data <= 64)
    mask_193_255 = (data > 192)
    mask_65_192 = (~(mask_0_64 | mask_193_255))

    mat += (p0 + (p25 - p0) / 64. * data) * mask_0_64.astype(numpy.float32)
    mat += (p25 + (p75 - p25) / 128. * (data - 64)) * mask_65_192.astype(numpy.float32)
    mat += (p75 + (p100 - p75) / 63. * (data - 192)) * mask_193_255.astype(numpy.float32)

    return mat.T  # transpose! col-major -> row-major,


def read_compressed_matrix_shape(fd):
    """ Read a compressed matrix,
        see: https://github.com/kaldi-asr/kaldi/blob/master/src/matrix/compressed-matrix.h
        methods: CompressedMatrix::Read(...), CompressedMatrix::CopyToMat(...),
    """

    # Format of header 'struct',
    global_header = numpy.dtype([('minvalue', 'float32'), ('range', 'float32'), ('num_rows', 'int32'),
                                 ('num_cols', 'int32')])  # member '.format' is not written,
    per_col_header = numpy.dtype([('percentile_0', 'uint16'), ('percentile_25', 'uint16'), ('percentile_75', 'uint16'),
                                  ('percentile_100', 'uint16')])

    # Read global header,
    globmin, globrange, rows, cols = numpy.frombuffer(fd.read(16), dtype=global_header, count=1)[0]

    return rows, cols


def read_matrix(f, dtype=numpy.float32):
    header = f.read(2).decode('utf-8')
    if header != "\0B":
        raise ValueError("Binary mode header ('\0B') not found when attempting to read a matrix.")
    format = read_string(f)

    if format == "DM":
        nRows = read_integer(f)
        nCols = read_integer(f)

        data = struct.unpack("<%dd" % (nRows * nCols), f.read(nRows * nCols * 8))
        data = numpy.array(data, dtype=dtype)
        return data.reshape(nRows, nCols)

    elif format == "FM":
        nRows = read_integer(f)
        nCols = read_integer(f)

        data = struct.unpack("<%df" % (nRows * nCols), f.read(nRows * nCols * 4))
        data = numpy.array(data, dtype=dtype)
        return data.reshape(nRows, nCols)

    elif format == "CM":
        data = read_compressed_matrix(f, format)
        return data

    else:
        raise ValueError(
            "Unknown matrix format '%s' encountered while reading; currently supported formats are DM (float64) and FM (float32)." % format)

def read_matrix_format(f):

    header = f.read(2).decode('utf-8')
    if header != "\0B":
        raise ValueError("Binary mode header ('\0B') not found when attempting to read a matrix.")

    format = read_string(f)

    return format


def read_matrix_shape(f):
    header = f.read(2).decode('utf-8')
    if header != "\0B":
        raise ValueError("Binary mode header ('\0B') not found when attempting to read a matrix.")

    format = read_string(f)

    # for compressed shape
    if format == 'CM':
        return read_compressed_matrix_shape(f)

    # for non compression shape
    nRows = read_integer(f)
    nCols = read_integer(f)

    if format == "DM":
        f.seek(nRows * nCols * 8, os.SEEK_CUR)
    elif format == "FM":
        f.seek(nRows * nCols * 4, os.SEEK_CUR)
    else:
        raise ValueError(
            "Unknown matrix format '%s' encountered while reading; currently supported formats are DM (float64) and FM (float32)." % format)
    return nRows, nCols


def write_string(f, s):
    return f.write((s + " ").encode('utf-8'))


def write_integer(f, a):
    s = struct.pack("<i", a)
    return f.write(chr(len(s)).encode('utf-8') + s)


def write_matrix(f, data):
    cnt = 0
    cnt += f.write('\0B'.encode('utf-8'))  # Binary data header
    if str(data.dtype) == "float64":
        cnt += write_string(f, "DM")
        cnt += write_integer(f, data.shape[0])
        cnt += write_integer(f, data.shape[1])
        cnt += f.write(struct.pack("<%dd" % data.size, *data.ravel()))

    elif str(data.dtype) == "float32":
        cnt += write_string(f, "FM")
        cnt += write_integer(f, data.shape[0])
        cnt += write_integer(f, data.shape[1])
        cnt += f.write(struct.pack("<%df" % data.size, *data.ravel()))
    else:
        raise ValueError(
            "Unsupported matrix format '%s' for writing; currently supported formats are float64 and float32." % str(
                data.dtype))

    return cnt

def read_matrix_by_offset(arkfile, offset, dtype=numpy.float32):
    with smart_open(arkfile, "rb") as g:
        g.seek(offset)
        feature = read_matrix(g, dtype)
    return feature


def read_scp_offset(filename, limit=numpy.inf):

    utt_ids = []
    ark_files = []
    offsets = []

    with smart_open(filename, "r") as f:
        for line in f:
            utt_id, pointer = line.strip().split()
            p = pointer.rfind(":")
            arkfile, offset = pointer[:p], int(pointer[p + 1:])
            ark_files.append(arkfile)
            offsets.append(offset)
            utt_ids.append(utt_id)

            if len(utt_ids) == limit: break
    return utt_ids, ark_files, offsets


from allosaurus.model import get_model_path

class Reporter:

    def __init__(self, train_config):
        self.train_config = train_config

        self.model_path = get_model_path(train_config.new_model)

        # whether write into std
        self.verbose = train_config.verbose

        # log file
        self.log_file = None

        self.open()

    def open(self):
        # whether write into log file
        self.log_file = None
        if self.train_config.log != 'none':
            self.log_file = open(self.model_path / 'log.txt', 'w', encoding='utf-8')

    def close(self):
        if self.log_file:
            self.log_file.close()

    def write(self, message):

        if self.verbose:
            print(message)

        if self.log_file:
            self.log_file.write(message+'\n')


import argparse
from pathlib import Path
from allosaurus.model import copy_model
from allosaurus.am.factory import transfer_am
from allosaurus.am.trainer import Trainer
from allosaurus.am.loader import read_loader

if __name__ == '__main__':

    parser = argparse.ArgumentParser("fine-tune an existing model to your target dataset")

    # required options
    parser.add_argument('--pretrained_model', required=True, type=str, help='the pretrained model id which you want to start with' )
    parser.add_argument('--new_model',        required=True, type=str, help='the new fine-tuned model id, this id show in your model list and will be available later for your inference')
    parser.add_argument('--path',             required=True, type=str, help='the data path, it should contain train directory and validate directory')
    parser.add_argument('--lang',             required=True, type=str, help='the language id of your target dataset')
    parser.add_argument('--device_id',        required=True, type=int, help='gpu cuda_device_id. use -1 if you do not have gpu')

    # non required options
    parser.add_argument('--batch_frame_size', type=int,   default=6000,  help='this indicates how many frame in each batch, if you get any memory related errors, please use a lower value for this size')
    parser.add_argument('--criterion',        type=str,   default='ctc', choices=['ctc'], help='criterion, only ctc now')
    parser.add_argument('--optimizer',        type=str,   default='sgd', choices=['sgd'], help='optimizer, only sgd now')
    parser.add_argument('--lr',               type=float, default=0.01,  help='learning rate')
    parser.add_argument('--grad_clip',        type=float, default=5.0,   help='grad clipping')
    parser.add_argument('--epoch',            type=int,   default=10,    help='number of epoch to run')
    parser.add_argument('--log',              type=str,   default='none',help='file to store training logs. do not save if none')
    parser.add_argument('--verbose',          type=bool,  default=True,  help='print all training logs on stdout')
    parser.add_argument('--report_per_batch', type=int,   default=10,    help='report training stats every N epoch')

    train_config = parser.parse_args()

    # prepare training and validating loaders
    data_path = Path(train_config.path)
    train_loader = read_loader(data_path / 'train', train_config)
    validate_loader = read_loader(data_path / 'validate', train_config)

    # initialize the target model path with the old model
    copy_model(train_config.pretrained_model, train_config.new_model)

    # setup the target model path and create model
    model = transfer_am(train_config)

    # setup trainer
    trainer = Trainer(model, train_config)

    # start training
    trainer.train(train_loader, validate_loader)

    # close datasets and loaders
    train_loader.close()
    validate_loader.close()


import torch
from collections import OrderedDict
import numpy as np

def torch_load(model, path, device_id, unit_mask=None):
    """Load torch model states.

    Args:
        path (str): Model path or snapshot file path to be loaded.
        model (torch.nn.Module): Torch model.
        device_id (int): gpu id (-1 indicates cpu only)

    """

    if device_id >= 0:
        model_state_dict = torch.load(str(path),map_location=torch.device(f'cuda:{device_id}'))
    else:
        model_state_dict = torch.load(str(path), map_location=torch.device('cpu'))

    new_state_dict = OrderedDict()
    for k, v in model_state_dict.items():

        # no need for lang specific layer in inference model
        if k.startswith('allophone_layer_dict'):
            continue

        if k.startswith('module.'):
            name = k[7:] # remove `module.`
        else:
            name = k

        # remap the phone_layer for fine-tuning
        # it will remap phone_layer.weight and phone_layer.bias
        if k.startswith('phone_layer'):
            if unit_mask is not None:
                phone_size = len(unit_mask.target_unit)

                if len(v.shape) == 2:
                    # for weight

                    hidden_size = v.shape[1]
                    new_v = v.new(phone_size, hidden_size)
                else:
                    # for bias

                    assert len(v.shape) == 1, 'phone_layer shape is either 2 or 1'
                    new_v = v.new(phone_size)

                for domain_phone_id, target_phone_id in unit_mask.unit_map.items():
                    new_v[target_phone_id] = v[domain_phone_id]

                v = new_v

        new_state_dict[name] = v

    if hasattr(model, 'module'):
        model.module.load_state_dict(new_state_dict)
    else:
        model.load_state_dict(new_state_dict)

    if device_id >= 0:
        model = model.cuda(device_id)

    del model_state_dict, new_state_dict


def torch_save(model, path):
    """Save torch model states.

    Args:
        path (str): Model path to be saved.
        model (torch.nn.Module): Torch model.

    """
    path = str(path)
    if hasattr(model, 'module'):
        torch.save(model.module.state_dict(), path)
    else:
        torch.save(model.state_dict(), path)


def apply_to_tensor(f, sample):
    if len(sample) == 0:
        return {}

    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, dict):
            return {
                key: _apply(value)
                for key, value in x.items()
            }
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        elif isinstance(x, tuple):
            return [_apply(x) for x in x]
        else:
            return x

    return _apply(sample)

def apply_to_ndarray(f, sample):
    if len(sample) == 0:
        return {}

    def _apply(x):
        if isinstance(x, np.ndarray):
            return f(x)
        elif isinstance(x, dict):
            return {
                key: _apply(value)
                for key, value in x.items()
            }
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        elif isinstance(x, tuple):
            return [_apply(x) for x in x]
        else:
            return x

    return _apply(sample)


def tensor_to_cuda(sample, device_id=0):

    def _move_to_cuda(tensor):
        return tensor.to(device_id)

    return apply_to_tensor(_move_to_cuda, sample)

def ndarray_to_tensor(sample):

    def _move_to_tensor(dnarray):
        return torch.from_numpy(dnarray)

    return apply_to_ndarray(_move_to_tensor, sample)

def move_to_tensor(sample, device_id=-1):
    """
    move numpy array to torch tensor

    :param sample:
    :param device_id: -1 means cpu, other means gpu device_id
    :return:
    """

    sample = ndarray_to_tensor(sample)

    # move to cuda if device_id provided
    if device_id >= 0:
        sample = tensor_to_cuda(sample, device_id)

    return sample

def move_to_ndarray(sample):

    if sample.is_cuda:
        sample = sample.cpu()

    return sample.data.numpy()



from setuptools import setup,find_packages

setup(
   name='allosaurus',
   version='0.5.0',
   description='a multilingual phone recognizer',
   author='Xinjian Li',
   author_email='xinjianl@cs.cmu.edu',
   url="https://github.com/xinjli/allosaurus",
   packages=find_packages(),
   install_requires=[
      'scipy',
      'numpy',
      'resampy',
      'panphon',
      'torch',
      'editdistance',
   ]
)



import unittest
from pathlib import Path
from allosaurus.app import read_recognizer

class TestRecognition(unittest.TestCase):

    def test_latest_nonempty(self):
        audio_file = Path(__file__).parent.parent / 'sample.wav'
        model = read_recognizer('latest')
        results = model.recognize(audio_file)
        self.assertTrue(len(results) > 0)

    def test_eng_nonempty(self):
        audio_file = Path(__file__).parent.parent / 'sample.wav'
        model = read_recognizer('latest')
        results = model.recognize(audio_file, 'eng')
        self.assertTrue(len(results) > 0)

if __name__ == '__main__':
    unittest.main()


from .articulatory import *
from .unit import *
from pathlib import Path

def read_prior(prior_path):

    prior = {}

    for i, line in open(str(prior_path), 'r', encoding='utf-8'):
        unit, prob = line.split()

        if i == 0:
            assert unit == '<blk>', 'first element should be blank'

        prior[unit] = np.log(prob)

    return prior



class UnitMask:

    def __init__(self, domain_unit, target_unit, approximation=False, inference_config=None):
        """
        MaskUnit provides interface to mask phones

        :param domain_unit: all available units (phones)
        :param target_unit: usually a subset of domain_unit
        """

        self.inference_config = inference_config

        self.domain_unit = domain_unit
        self.target_unit = target_unit

        # whether or not to use articulatory feature to map unseen units
        self.approximation = approximation

        # available index in all_unit
        self.valid_mask = set()

        # invalid index in all_unit
        self.invalid_mask = set()

        # index mapping from all_unit to target_unit
        self.unit_map = dict()

        # prior
        self.prior = np.zeros(len(self.domain_unit), dtype=np.float32)

        if self.inference_config and self.inference_config.prior and Path(self.inference_config.prior).exists():
            self.create_prior()

        # mask
        self.create_mask()

        if self.approximation:
            self.articulatory = Articulatory()
            self.approxmiate_phone()

        # create a mask for masking numpy array
        self.invalid_index_mask = sorted(list(self.invalid_mask))


    def __str__(self):
        return '<UnitMask: valid phone: ' + str(len(self.valid_mask)) + ', invalid phone: '+ str(len(self.invalid_mask))+'>'

    def __repr__(self):
        return self.__str__()


    def create_mask(self):

        # invalidate all phones first
        self.invalid_mask = set(range(1, len(self.domain_unit)))

        # <blank> is valid phone
        self.valid_mask.add(0)

        # <blank> should be mapped to <blank>
        self.unit_map[0] = 0

        # register all valid phones
        for target_idx, target_phone in self.target_unit.id_to_unit.items():
            if target_phone in self.domain_unit:
                domain_idx = self.domain_unit.get_id(target_phone)

                # this domain_idx is available
                self.valid_mask.add(domain_idx)

                # remove the domain_idx from the invalid_mask
                self.invalid_mask -= { domain_idx }

                # register this domain idx -> target_idx
                self.unit_map[domain_idx] = target_idx

    def create_prior(self):

        for line in open(str(self.inference_config.prior), 'r', encoding='utf-8'):
            phone, logprob = line.strip().split()

            if phone in self.domain_unit:
                domain_idx = self.domain_unit.get_id(phone)
                self.prior[domain_idx] = logprob

    def approxmiate_phone(self):

        # register all valid phones
        for target_idx, target_phone in self.target_unit.id_to_unit.items():
            if target_phone not in self.domain_unit:

                # find the most similar phone from the invalid set
                max_domain_idx = -1
                max_domain_score = -10000

                for domain_idx in self.invalid_mask:

                    domain_phone = self.domain_unit.get_unit(domain_idx)

                    #print("domain ", domain_phone, " target ", target_phone)
                    score = self.articulatory.similarity(domain_phone, target_phone)

                    if score >= max_domain_score:
                        max_domain_score = score
                        max_domain_idx = domain_idx


                assert max_domain_idx not in self.valid_mask
                assert max_domain_idx != -1

                # map max_domain_idx to target_idx
                self.invalid_mask -= { max_domain_idx }
                self.valid_mask.add(max_domain_idx)

                #print("target phone", target_phone, ' mapped to idx ', max_domain_idx)

                self.unit_map[max_domain_idx] = target_idx


    def print_maps(self):

        for domain_id, target_id in self.unit_map.items():
            print(self.domain_unit.get_unit(domain_id)+' --> '+self.target_unit.get_unit(target_id))


    def mask_logits(self, logits):
        # mask inavailable logits

        # apply mask
        logits[:,self.invalid_index_mask] = -100000000.0

        # apply prior
        logits += self.prior

        return logits


    def get_units(self, ids):
        """
        get unit from ids

        :param ids: elem_id list
        :return: a list of unit
        """

        unit_lst = []

        for idx in ids:

            assert idx in self.unit_map

            target_idx = self.unit_map[idx]

            unit_lst.append(self.target_unit.get_unit(target_idx))

        return unit_lst


from allosaurus.pm.mfcc import MFCC
import json
from argparse import Namespace

def read_pm(model_path, inference_config):
    """
    read feature extraction model

    :param pm_config:
    :return:
    """

    pm_config = Namespace(**json.load(open(str(model_path / 'pm_config.json'))))

    assert pm_config.model == 'mfcc_hires', 'only mfcc_hires is supported for allosaurus now'
    assert pm_config.backend == 'numpy', 'only numpy backend is supported for allosaurus now'

    model = MFCC(pm_config)
    return model


from allosaurus.app import read_recognizer
from allosaurus.model import get_all_models, resolve_model_name
from allosaurus.bin.download_model import download_model
from pathlib import Path
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('Allosaurus phone recognizer')
    parser.add_argument('-d', '--device_id', type=int, default=-1, help='specify cuda device id to use, -1 means no cuda and will use cpu for inference')
    parser.add_argument('-m', '--model', type=str, default='latest', help='specify which model to use. default is to use the latest local model')
    parser.add_argument('-l', '--lang', type=str,  default='ipa',help='specify which language inventory to use for recognition. default is to use all phone inventory')
    parser.add_argument('-i', '--input', type=str, required=True, help='specify your input wav file/directory')
    parser.add_argument('-o', '--output', type=str, default='stdout', help='specify output file. the default will be stdout')
    parser.add_argument('-k', '--topk', type=int, default=1, help='output k phone for each emitting frame')
    parser.add_argument('-p', '--prior', type=str, required=False, default=None, help='supply prior to adjust phone predictions')
    parser.add_argument('-a', '--approximate', type=bool, default=False, help='the phone inventory can still hardly to cover all phones. You can use turn on this flag to map missing phones to other similar phones to recognize. The similarity is measured with phonological features')

    args = parser.parse_args()

    # download specified model automatically if no model exists
    if len(get_all_models()) == 0:
        download_model('latest')

    # resolve model's name
    model_name = resolve_model_name(args.model)
    if model_name == "none":
        print("Model ", model_name, " does not exist. Please download this model or use an existing model in list_model")
        exit(0)

    args.model = model_name

    # create recognizer
    recognizer = read_recognizer(args)

    # output file descriptor
    output_fd = None
    if args.output != 'stdout':
        output_fd = open(args.output, 'w', encoding='utf-8')

    # input file/path
    input_path = Path(args.input)

    if input_path.is_dir():
        wav_list = sorted(list(input_path.glob('*.wav')))
        for wav_path in wav_list:
            phones = recognizer.recognize(str(wav_path), args.lang)

            # save to file or print to stdout
            if output_fd:
                output_fd.write(wav_path.name+' '+phones+'\n')
            else:
                print(wav_path.name+' '+phones)

    else:

        # check file format
        assert args.input.endswith('.wav'), " Error: Please use a wav file. other audio files can be converted to wav by sox"

        # run inference
        phones = recognizer.recognize(args.input, args.lang, args.topk)

        if output_fd:
            output_fd.write(phones+'\n')
        else:
            print(phones)

    if output_fd:
        output_fd.close()